/*
Sample NetWin Authentication routine, high speed simple text file based
authentication

  Version see below

Command line options:
	nwauth  -set username[@domain] password INFO
	nwauth  -mod username[@domain] INFO
	nwauth  -check username password
	nwauth  -lookup username
	nwauth  -del username
	nwauth  -search string
	nwauth  -encrypt user/all
	nwauth -dodrop   (return a made up drop path /testdrop/username )

Normal processing options when used as external authent module for dsmtp/dpop:
	exit
	+OK
	set user password fwd="fred" groups="adults" name="Mr \"Cool\" Smith"
	check username[@domain] password fromIPaddress
	+OK username@domain drop_file_path uid
	lookup username[@domain]
	+OK username@domain drop_file_path uid fwd="a@b" groups="adults"
	set user password aaa="bbb" ccc="ddd"
	+OK    user added/modified
	delete user
	+OK    user removed
	search string
	+DATA    user1 info
	+DATA    user2 info
	+DATA    user3 info
	+DATA    user4 info
	+OK

Normal processing commands
	set user password [INFO]
	set user (NULL) [INFO]   (change info but leave password unchanged)
	check username[@domain] password fromIPaddress
	lookup username[@domain]
	delete user
	search string

Works by using a file nwauth.txt in this format
	user:password:info
New entries are placed into file, nwauth.add, and once this
reaches 'rebuild_size' (3kbytes) nwauth.add is removed and
nwauth.txt re-written.

To compile as part of WAddUser you should define NOAUTHMAIN.

Changes History:  (add to BOTTOM and increment version number EVERYTIME!)
----------------------------------------------------------------

  TRW added: -log option as well as -debug, only log to file,
					dsmtp_path\nwauth.txt


	Regan Added :
		-Label EXTERNAL_LOG, which disables nwauth log routines and enables other
		log routines to be called ie routines in log.c etc...
		- if NOAUTHMAIN then myprintf are replaced with buf_print
          buf_clear and buf_get are companion routines
	Regan Changed :
		-Now prints +OK after recieving quit command.
		-ncpy is now static as conflicted when included with other .c files that also included
		 ncpy.

  
2.0b;
	29/4/99 for 2.5d TRW changed:
		- \n on end of +OK
		-
	5/5/99  still for 2.5d, TRW changed;
		- fflush after +ok in response to exit, for most other
		responses it is in the end of the do_command function.
2.0c
	11/5/99 for 2.5d TRW changed;
		- made db_check and search check that build_index and check_add
			had been done at least once (for wadduser as is CGI).
		- added -version command
2.0d
	17/5/99 for 2.5f TRW changed
		- added -size command line option
		- made log file get logged to dmail dir (or as set by path
		command).
		- added more error logging in search.
{2.0e}
   30/June/99 Stuart Added:
     - passwords with a '*' in front of them are treated as plain
	 passwords and won't use the crypt function.
	 -encrypt <user/all> will encrpyt a users password if it
		isn't already encrypted, sending all as the argument
		will make it encrypt all users passwords that are not
		already encrypted.
2.0f
	6 July 99  For dmail 2.5h TRW changed;
	- merged 2.0d with 2.0e to give 2.0f
	- added Lynden's -fallback_unix option
	- removed special search where match contained '@'
	- made file_config use arg_path when no dmail.conf on machine.
2.0g
	10 August 99
	- made logging always append
	- added lockfile handling during rebuilds
2.0h
	30 August 1999
	Changes:
	- When nwauth.txt is re-written nwauth.bck is
		created if it didn't previously exist.
	- added command options
			nwauth -test to see if users are missing
			nwauth -fix to see and fix missing users
2.0i
	10 September 1999
	- made always log user adds and modifies as well as deletes.
	- made wadduser log lines start 'WA'
2.0j
	11 November 1999
	- added -sleep x y command line option for testing so that
	every xth lookup nwauth sleeps for y seconds before responding.
2.0k
	18 November 1999  - TRW.
	- FIXED BUG: -path setting was not being set before build index
		was done.
2.0l (for 2.8a of DMAIL)
	26 November 1999 - TRW
	- added -drop_path <hash> <base>, where hash is one of 0,1,2
		being the DMail hash_spool settings and base is the
		base directory path for drop files on all domains.  NB: this
		is intended for testing only and NOT for live system use.

Fri 28/Jan/1999 - Chrisp
	- Added chmod calls to force 0660 mode on nwauth.* files
	- Added emsg when command takes more than 10 seconds.

2.0m
	22 Dec 1999 - CHRISP
	- Added -generate n command switch for creating test users
	- Fixed search so it uses myprintf

	18 Feb 2000 - TRW
	- added logging of full command received if -log command line
		option given
2.0n (dmail 2.8f)
	23 March 2000 - TRW
	- added -filtername,x,y option
2.0o (dmail 2.8i)
	16 May 2000 - TRW
	- added a while back, log when response time over 10 seconds.
	- added -logpass which causes the full received line to be logged including the 
		password.
	- disallow by default the following chars in the local part of usernames (in front of @) by default,
				BADCHARS_LOCALPART ",:|;[]()<>\"/\\"
	- disallow by default the following chars in any part of usernames by default,
				BADCHARS ":|\""
	- disallow highascii chars in any part of username, i.e. chars 0-31 and 127 up.

	- added -badchar <string-list>, where string-list specifies a list of 
			bad characters to override the default list with
	- added -badchar_localpart <string-list>, where string-list specifies a list of 
			bad characters to override the default localpart (before the @) list with
	-added 	-allowhighascii, to allow usernames with chars in the ranges 0-31 and >127.

	- BUG FIX: if encoded password (in nwauth.add) is less than 2 chars log then respond negatively (to stop 
		situation where "jeff:b:pass" in nwauth.add results in any password succeeding on a check))
	- BUG FIX: now responds to set command when adding a user fails! (unsure when this stopped being the case).	
	- more than one @ symbol no longer allowed in username by default
	- added -allowats which overrides default limitation of one @ symbol in username
	- now logs location of dmail.conf file that it is reading.
2.0p TRW 13 Jun 2000
	- made wadduser check that auth_set succeeds before showing 'added.htm' so that 
		 if username is invalid it does not report the user added ok.
2.0q TRW 9 August 2000
	- made error message when cannot access nwauth.add file reply -DEAD and give string error.
	- file locking adding in version previous to this so safe for NFS.
2.0r TRUE 11 August 2000
	- FIXED BUG: mod from command line was not passing the first argument of the 
		info fields (actually arg 2).
	- FIXED BUG: mod command in non-command line mode was overwriting password with 
		the first info field argument.
2.0s 28 AUG 2000 Regan
	- Added -from and -max params to the search function, -max causes 
	a maximum number of users to be returned. -from causes listing to 
	start from the x'th match.
2.0t
	- added #ifdef NEEDSTRLWR instead of strlwr on #ifndef WIN32
	- added rejection of usernames > 255 chars on set.
	- changed imsg on wrong password to show password given if -logpass option on.
2.0u RMH
	- Added wild_match function.
	- Search uses wild_match if search string contains a *
2.0v RMH
	- Search was behaving strangely and returning more than expected with certain 
	combinations of the -max and -from parameters.
2.0w	TRW 1 Feb 2000
	- made search do a count of entries found so response now has 'out of x' on the end.
2.0x RMH 5 Feb 2001
	- Added static int total_user_count to count total number of users so search didn't have to.
	TRW, added FALLBACK_UNIX define on -Dunix, if not able to use on some systems then 
	use -DNO_FALLBACK_UNIX option.
dmail version 30a1
	- added -fallback_unix_crypt (NT only) option which does UNIX crypt() function to check 
	password if other checks have failed - allows the use of UNIX /etc/password 
	entries on NT.
	nwauth VERSION now always same as current dmail_beta
	nwauth no longer builds with dmail on NT - so build after every change.
dmail version 30a2 (and stable 2.9 something) - CHRISP CHANGES
	Fixed fault with not re-reading add file after re-reading main index
	Fixed fault with not seeing changes to .add file after getting lock on main index
nwauth Version 4.0 
	Removed all buffer over run bugs.
	Add optional sha encoding and made default for windows and unix 

Version 4.0b
	Increased timeout if file not found and not ENOENT, and 
	improved logging.
Version 4.0c
	Fixed bug. in 4.0a,b


todo:
	When change or set command, add to '.add' file
	Make database file a command line option.

*/
/* definable options */
/* #define NOAUTHMAIN */
/* #define EXTERNAL_LOG*/
/* #define FALLBACK_PASSWORD*/


/*#define VERSION "version 2.0x"*/

#define VERSION "4.0c"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "ld_sha.h"

#ifndef WIN32
#include <unistd.h>
#ifndef HAVESTRLWR
char *strlwr(char *s);
#endif
#endif

#ifndef WIN32
#ifndef NO_FALLBACK_UNIX  /*use this define to turn off fallback option if necessary*/
#define FALLBACK_UNIX
#endif
#endif

#ifndef PASS_NOSHADOW
#ifndef WIN32
#define HAS_SHADOW
#endif
#endif

#ifdef WIN32
	int sleep(int x);
#	include <process.h>
	char *crypt(char *key, char *salt);
#	define getpid _getpid
#	define PATH_SEP '\\'
#	include "./fcrypt.h" /*copy from h:\src\fcrypt\*/
#else
#	define PATH_SEP '/'
	char *crypt(const char *key, const char *salt);
#	ifdef FALLBACK_UNIX
#		include <pwd.h>
#		include <sys/types.h>
#		ifdef HAS_SHADOW
#			include <shadow.h>
#			ifdef HAS_PWAUTH
#				include <shadow/pwauth.h>
				char *pw_encrypt (char *, char *);	/* To permit long shadow passwords */
#				define crypt pw_encrypt		/* for short passwords as well.    */
#			endif
#			ifndef PW_PPP
#				define PW_PPP PW_LOGIN
#			endif
#		endif
		struct passwd *pw;
		int chkPass (char *user, char *pass, struct passwd *pw);
#	endif
#endif
#include "nwauth.h"
#define REAL_LOCKING 1
#ifdef REAL_LOCKING
#include "flock.h"
#endif
int test_get_cmd(char *bf);

#ifdef WIN32
#define vsnprintf _vsnprintf
#define snprintf _snprintf
#endif

int db_check(char *user, char *pass, char *info);
char *auth_info(char *argv[], int i, int argc);
int auth_param(char *argv[], char *name, int argc);
int lib_date(char *fname);
char *value_encode(char *s);
int auth_rebuild(void);
void do_vers(void);
int auth_change_isdeleted(char *thisuser, FILE *f);
int auth_mod(char *user, char *info, char *xx);
#define MODE_PROT 0660

int auth_generate_fast(int n);
int auth_generate(int n);
int auth_write_backup(void);
char *str_encode(char *s);
char *file_config(char *fname);
char *file_dmail(void);
void check_add(void);
int free_index(void);
int makelist(char *filename);
int has_changed(void);
int do_command(char *s);
int build_index(char *reason);
int str_hash(char *s, int max);
char * db_pass(char *user);
void log_file(char *s);
void auth_fixup(char *user, char *pass, char *info);
static char *ncpy(char *dst, char *src, int len);
int auth_lib_exists(char *fname);
int auth_change_test(void);
int auth_mod_raw(char *user, char *pass, char *addinfo);
int auth_set_raw(char *user, char *pass, char *info);
int auth_del_raw(char *user, int silent);
int auth_lock(void);
int auth_unlock(void);

char *vdm(char *user);/*converted from dpop function for -drop_path*/
int do_hash(char *dp,int spool,char *uid);/*converted from dpop function for -drop_path*/

void filtername(char *name);
int findfilters(char *arg);
char *nwdrop_path(char *user);

int auth_changelog(char *user, char *pass, char *info);
void zfreeall(void);
void zstrfree(char *s);
char *zstrdup(char *s);
int token_split_quote(char *s, char *p[], char *sep, int tmax);
int username_badchar(char *user); /*check for bad characters username*/
static int add_file_reread;
int auth_unlock_real(void);
int auth_unlock_nfs(void);
int auth_lock_real(void);
int auth_lock_nfs(void);
static int isnfs;
static int dodrop = 0;
char *zstrstrnc(char *s1, char *s2) ;
int zstrncmpnc(char *s1, char *s2, int n) ;
int zstrcmpnc(char *s1, char *s2);
int auth_isme(FILE *f);
static int dossha=1;
int wild_match(char *wild,char *text);

#ifdef EXTERNAL_LOG
#	include "new_log.h"
#else
	void dmsg(char * arg_list, ...);
	void emsg(char * arg_list, ...);
#    define imsg dmsg
#endif

#ifdef NOAUTHMAIN
#	define myprintf buf_print
#else
#   define myprintf printf
#endif

#define BFSZ 1000
#ifndef FALSE
#	define FALSE 0
#	define TRUE (!FALSE)
#endif

#define BADCHARS ":|\"" /*  :|"  */
#define BADCHARS_LOCALPART ",:|;[]()<>\"/\\" /*  ,:|;[](),."/\  (spaces are allowed)*/
	
static int isfiltername;/*enables filtering of name given to replace character(s)*/
char filterfind;/*char to find*/
char filterreplace;/*replace found char with this, may be blank*/
static int isdebug;/*enables msgs to stdout on when not in slave mode*/
static int islog;/*enables msgs to be logged in any mode*/
static int isend;
static int isfallback_unix;/*must define FALLBACK_UNIX at compile time to use this*/
static int isfallback_unix_crypt; /*command line option, if set tries unix crypt on nt*/
static int issleep;/*sleep for y on every xth lookup*/
static int islogpass; /*log full log line given including password!*/
static int sleepx,sleepy,cmdx;
static int isdroppath,hash_spool;  /*makes nwauth respond with drop paths for users*/
static int allowhighascii;  /*turns off default setting to disallow high ascii when setting usernames.*/
static int allowats; /*overrides default limitation of one @ symbol in username*/
static char hash_base[BFSZ];
static char badchar[BFSZ];
static char badchar_localpart[BFSZ];
static int pid;
static int encrypt_on = FALSE;
static int done_check_add;
static int done_build_index;
#define DFLT_REBUILD_SIZE 3000
static int rebuild_size;

static char *buffer = NULL;

static int total_user_count = 0;

int f_size(FILE *f);
int auth_init(void)
{
	return build_index("init");
}
#ifdef WIN32
#undef chmod
int chmod(const char *s, int x)
{
	return 0;
}
#endif
static void (*nwauth_print)(char *s);
void nwauth_setprint(void (*pr)(char *s))
{
	nwauth_print = pr;
}
int buf_print(char *arg_list, ...)
{
	va_list arg_ptr;
    char *format, *p;
  	char text[4512];
    size_t size = 0, new_size = 0;

	va_start(arg_ptr, arg_list);
	format = arg_list;
	vsnprintf(text,4000, format, arg_ptr);
	if (nwauth_print!=NULL) {
		nwauth_print(text);
		return strlen(text);
	}

	if (buffer) size = strlen(buffer);
	new_size = size + strlen(text);

	buffer = realloc(buffer, new_size + 2);
	p = buffer + size;

	strcpy(p, text);
	return size;
}
int buf_clear(void)
{
	if (buffer) {free(buffer); buffer = NULL;}
	return 1;
}
char *buf_get(void)
{
	if (buffer) return buffer;
	else return "";
}
static int istest,isfix;
static char arg_path[BFSZ];
static 	int test_mode = FALSE;
#ifndef NOAUTHMAIN

int main(int argc, char *argv[])
{
	char bf[BFSZ];
	int t,i;
	time_t xstart;
	t = 1;

	srand(time(NULL));

	isfallback_unix=FALSE;
	isfallback_unix_crypt=FALSE;

	for (i=1;i<argc;i++) {
		if (strcmp(argv[i], "-encrypt")==0)
		     encrypt_on = TRUE;
		else if (strcmp(argv[i],"-debug")==0) /*WARNING debug may print to stdout*/
			isdebug = TRUE;
		else if (strcmp(argv[i],"-dodrop")==0) /*WARNING debug may print to stdout*/
			dodrop = TRUE;
		else if (strncmp(argv[i],"-filtername",11)==0) {
			if (findfilters(argv[i])) isfiltername=TRUE;
		}
#ifdef FALLBACK_UNIX
		else if (strcmp(argv[i],"-fallback_unix")==0)
			isfallback_unix = TRUE;
#endif
#ifdef WIN32
		else if (strcmp(argv[i],"-fallback_unix_crypt")==0)
			isfallback_unix_crypt = TRUE;
#endif
		else if (strcmp(argv[i],"-test_con")==0) {	test_mode = TRUE;  }
		else if (strcmp(argv[i],"-test")==0) {	istest = TRUE; }
		else if (strcmp(argv[i],"-fix")==0) { 	istest = TRUE;  isfix = TRUE;}
		else if (strcmp(argv[i],"-log")==0)
			islog = TRUE;
		else if (strcmp(argv[i],"-nfs")==0)
			isnfs = TRUE;
		else if (strcmp(argv[i],"-old_crypt")==0)
			dossha = FALSE;
		else if (strcmp(argv[i],"-version")==0){
			do_vers();
			return 0;
		}
		else if (strcmp(argv[i],"-sleep")==0) /*WARNING issleep causes out of sync responses!*/
			issleep = TRUE;
		else if (strcmp(argv[i],"-logpass")==0) 
			islogpass = TRUE;
		else if (strcmp(argv[i],"-allowhighascii")==0) 
			allowhighascii = TRUE;
		else if (strcmp(argv[i],"-allowats")==0) 
			allowats = TRUE;
		else if (strcmp(argv[i],"-path")==0)  {
			if ((i+1)<argc) ncpy(arg_path,argv[i+1],100);
			else goto badp;
		}
	}
	/*for (i=1;argc>i;i++)
	{
		if (strcmp(argv[i], "-encrypt")==0)
		     encrypt_on = TRUE;
	}*/
	pid=getpid();
	if (isdebug)
		dmsg("**NWAuth %s Started in debug mode, pid=%d, reading dmail.conf {%s},exe/log_path={%s}\n",VERSION,pid,file_dmail(),file_config(""));
	if (issleep)
		emsg("**NWAuth %s Started in SLEEP testing mode, pid=%d, reading dmail.conf {%s}, exe/log_path={%s}\n",VERSION,pid,file_dmail(),file_config(""));
	if (islog)
		dmsg("**NWAuth %s Started in logging mode, pid=%d, reading dmail.conf {%s}, exe/log_path={%s}\n",VERSION,pid,file_dmail(),file_config(""));

	if (!build_index("start")) return 1;
	check_add(); /* Check for new/changed entries */
	if (istest) {
		printf("Doing tests\n");
		auth_change_test();
		return 0;
	}

again:
	/*TRW; if (argc>t) {
		if (strcmp(argv[t],"-path")==0) {if (t+1>=argc) goto badp; strcpy(arg_path,argv[t+1]); t+=2; goto again;}
		else if (strcmp(argv[t],"-debug")==0) {
			t++;
			isdebug = TRUE;
		}
	}

	if (!build_index()) return 1;
	check_add(); /* Check for new/changed entries */
	if (argc>t) {
		if (strcmp(argv[t],"-add")==0)	{if (t+3>=argc)  auth_set(argv[t+1],argv[t+2],auth_info(argv,t+3,argc));}
		else if (strcmp(argv[t],"-set")==0) {if (t+2>=argc) goto badp; auth_set(argv[t+1],argv[t+2],auth_info(argv,t+3,argc));}
		else if (strcmp(argv[t],"-generate")==0) {auth_generate(atoi(argv[t+1]));}
		else if (strcmp(argv[t],"-generate_fast")==0) {auth_generate_fast(atoi(argv[t+1]));}
		/*else if (strcmp(argv[t],"-mod")==0) {if (t+1>=argc) goto badp; auth_set(argv[t+1],"(NULL)",auth_info(argv,t+3,argc));}*/
		else if (strcmp(argv[t],"-mod")==0) {if (t+1>=argc) goto badp; auth_set(argv[t+1],"(NULL)",auth_info(argv,t+2,argc));}
		else if (strcmp(argv[t],"-del")==0) {if (t+1>=argc) goto badp; auth_del(argv[t+1],FALSE);}
		else if (strcmp(argv[t],"-encrypt")==0){if (t+1>=argc) goto badp;encrypt_users(argv[t+1]);}
		else if (strcmp(argv[t],"-makelist")==0){if (t+1>=argc) goto badp;makelist(argv[t+1]);}
		else if (strcmp(argv[t],"-check")==0) {if (t+2>=argc) goto badp; auth_check(argv[t+1],argv[t+2]);}
		else if (strcmp(argv[t],"-lookup")==0) {if (t+1>=argc) goto badp; auth_lookup(argv[t+1]);}
		else if (strcmp(argv[t],"-search")==0) {if (t+1>=argc) goto badp; auth_search(argv[t+1], auth_param(argv,"-from",argc), auth_param(argv,"-max",argc));}
		else if (strcmp(argv[t],"-path")==0) {if (t+1>=argc) goto badp; ncpy(arg_path,argv[t+1],100); t+=2; goto again;}
		else if (strcmp(argv[t],"-old_crypt")==0) {t++; dossha = FALSE; goto again;}
		else if (strcmp(argv[t],"-size")==0) {if (t+1>=argc) goto badp; rebuild_size=atoi(argv[t+1]); t+=2; goto again;}
		else if (strcmp(argv[t],"-badchar")==0) {if (t+1>=argc) goto badp; strncpy(badchar,argv[t+1],BFSZ-1); t+=2; goto again;}
		else if (strcmp(argv[t],"-badchar_localpart")==0) {if (t+1>=argc) goto badp; strncpy(badchar_localpart,argv[t+1],BFSZ-1); t+=2; goto again;}
		else if (strcmp(argv[t],"-log")==0) {t++;  goto again;}
		else if (strcmp(argv[t],"-dodrop")==0) {t++;  goto again;}
		else if (strcmp(argv[t],"-nfs")==0) {t++;  goto again;}
		else if (strcmp(argv[t],"-test_con")==0) {t++;  goto again;}
		else if (strcmp(argv[t],"-logpass")==0) {t++;  goto again;}
		else if (strcmp(argv[t],"-allowats")==0) {t++;  goto again;}
		else if (strcmp(argv[t],"-allowhighascii")==0) {t++; goto again;}
		else if (strcmp(argv[t],"-debug")==0) {t++; goto again;}
		else if (strncmp(argv[t],"-filtername",11)==0) {t++; goto again;}
		else if (strcmp(argv[t],"-sleep")==0) {if (t+1>=argc) goto badp; sleepx=atoi(argv[t+1]); sleepy=atoi(argv[t+2]); t+=3; goto again;}
		else if (strcmp(argv[t],"-drop_path")==0) {if (t+2>=argc) goto badp; isdroppath=TRUE; hash_spool=atoi(argv[t+1]); ncpy(hash_base,argv[t+2],200); t+=3; goto again;}
		else if (isfallback_unix) {if (strcmp(argv[t],"-fallback_unix")==0) {t++; goto again;}}
		else if (isfallback_unix_crypt) {if (strcmp(argv[t],"-fallback_unix_crypt")==0) {t++; goto again;}}
		else if (strcmp(argv[t],"-help")==0) {
			myprintf("Usage:\n");
			myprintf("\tnwauth -set user password var1=value ...\n");
			myprintf("\tnwauth -mod user var1=value ...\n");
			myprintf("\tnwauth -del user\n");
			myprintf("\tnwauth -check user password\n");
			myprintf("\tnwauth -lookup user\n");
			myprintf("\tnwauth -search string\n");
			myprintf("\tnwauth -search string [-max n] (don't return more than n entries)\n");
			myprintf("\tnwauth -search string [-from n] (start returning from n'th \n");
			myprintf("\t\tmatching entry)\n");
			myprintf("\tnwauth -encrypt user/all\n");
			myprintf("\tnwauth -generate n (creates accounters test0..testn passwords test)\n");
			myprintf("\tGive no switches to run in SLAVE MODE\n\n");
			myprintf("\tOptional switches for any mode:\n");
			myprintf("\tnwauth -size x -... (sets max size of nwauth.add)\n");
			myprintf("\tnwauth -log -... (turns on logging to nwauth.log)\n");
			myprintf("\tnwauth -sleep x y (testing option, sleep for y seconds first and every\n");
			myprintf("\t\txth response)\n");
			myprintf("\tnwauth drop_path <hash> <base> (testing option, give drop path using\n"); 
			myprintf("\t\tdpop hashing, hash=0,1,2, base is common drop path)\n");
			myprintf("\tnwauth -logpass -... (testing option, log full command as given (\n");
			myprintf("\t\tincludes password!))\n");
			myprintf("\tnwauth -badchar <string> -... (overrides default bad character list\n"); 
			myprintf("\t\twith the string)\n");
			myprintf("\tnwauth -badchar_localpart <string> -... (overrides default localpart (\n");
			myprintf("\t\tbefore the @) bad character list with the string)\n");
			myprintf("\tnwauth -allowats -... (disables check for more than one '@' symbol in\n"); 
			myprintf("\t\tusername)\n");
			myprintf("\tnwauth -allowhighascii -... (disables check for high and low ascii in\n"); 
			myprintf("\t\tusername)\n");


		}
		else emsg("ERROR: invalid switch {%s}\n",argv[t]);/*TRW added 2.0f*/
		return 0;
badp:		myprintf("Wrong number of paramters\n");
			fflush(stdout);
		return 0;
	}

	if (isdebug) {dmsg("isdebug so not allowing nwauth to run in slave mode\n"); isend=TRUE;}

	for (;!isend;) {
		if (test_mode) test_get_cmd(bf);
		else if (fgets(bf,BFSZ-1,stdin)==NULL) break;
		xstart = time(NULL);
		/* Check if file changes (once a second at most) */
		if (issleep){
			cmdx++;
			/*printf("%d %d\n",cmdx,cmdx % sleepy); /*xxxxx*/
			if (cmdx==1 || !(cmdx % sleepx))
				sleep(sleepy);

		}
		if (has_changed()) build_index("changed");
		check_add(); /* Check for new/changed entries */
		/*if (islog||isdebug) dmsg("<-- (%.200s)\n",bf);*/
		do_command(bf);
		if (time(NULL)-xstart>10) {
			emsg("ERROR: Command took %d seconds: %.80s\n",time(NULL)-xstart,bf);
		}
	}
	myprintf("+OK\n");
	fflush(stdout);
	return 0;
}
#endif
void do_vers(void)
{
	myprintf("+OK nwauth %s\n",VERSION);
	fflush(stdout);
}
char *auth_info(char *argv[], int i, int argc)
{
	static char info[BFSZ];
	strcpy(info,"");
	for (;i<argc;i++) {
		strcat(info,argv[i]);
		strcat(info," ");
	}
	return info;
}
/* NOTE i=3 skips exe name, "search" and "*" */
int auth_param(char *argv[], char *name, int argc) {
	if ( name ) {
		int i;

		for (i=3;i<argc;i++) {
			if (strcmp(name, argv[i])==0 && i+1 <= argc-1)
				return atoi(argv[i+1]);
		}
	}
	return 0;
}

/*
void auth_silent(int mode)
{
	silent = mode;
}
int auth_silent_get(void)
{
	return silent;
}
*/
int auth_testcrypt(void)
{
	int i;
	char bf[BFSZ];

	for (i=0; i<1000; i++) {
		sprintf(bf,"test%d",i);
		auth_set(bf,bf,"");
		auth_check(bf,bf);
	}
	return TRUE;
}
int do_command(char *bf)
{
	char *s;
	char cmd[BFSZ];
	char p1[BFSZ];
	char p2[BFSZ];
	char p3[BFSZ];

	if (islogpass)
		emsg("recvd: (%.200s)\n",bf);
	strcpy(cmd,"nocommand");
	s = strtok(bf," \n\r"); if (s==NULL) goto failed;
	ncpy(cmd,s,BFSZ-1);
	strcpy(p1,"");
	strcpy(p2,"");
	strcpy(p3,"");
	s = strtok(NULL," \n\r");
	if (s!=NULL) ncpy(p1,s,BFSZ-1);
	if (s!=NULL) s = strtok(NULL," \n\r");
	if (s!=NULL) ncpy(p2,s,BFSZ-1);
	if (s!=NULL) s = strtok(NULL,"\n\r");
	if (s!=NULL) ncpy(p3,s,BFSZ-1);
		
	strlwr(cmd);
	if (strcmp(cmd,"check")==0) auth_check(p1,p2);
	else if (strcmp(cmd,"lookup")==0) auth_lookup(p1);
	else if (strcmp(cmd,"set")==0) auth_set(p1,p2,p3);
	/*else if (strcmp(cmd,"mod")==0) auth_mod(p1,p2,p3);*/
	else if (strcmp(cmd,"mod")==0) auth_mod(p1,"(NULL)",p3);
	else if (strcmp(cmd,"del")==0) auth_del(p1, FALSE);
	else if (strcmp(cmd,"testcrypt")==0) auth_testcrypt();
	else if (strcmp(cmd,"search")==0) {
		int from=0,max=0;
		char *ptr = p3;

		if ( strcmp(p2, "-from") == 0 )
			from = atoi(ptr);
		else if ( strcmp(p2, "-max") == 0 )
			max = atoi(ptr);
		
		ptr = strstr(p3, "-from");
		if ( ptr ) {
			while ( *ptr != ' ' && *ptr != '\t' )
				ptr++;
			from = atoi(ptr);
		}
		ptr = strstr(p3, "-max");
		if ( ptr ) {
			while ( *ptr != ' ' && *ptr != '\t' )
				ptr++;
			max = atoi(ptr);
		}

		auth_search(p1, from, max);
	}
	else if (strcmp(cmd,"quit")==0) isend = TRUE;
	else if (strcmp(cmd,"exit")==0) isend = TRUE;
	else if (strcmp(cmd,"version")==0) {do_vers();}
	else if (strcmp(cmd,"help")==0) {
		myprintf("+DATA Valid commands\n");
		myprintf("+DATA \tlookup username\n");
		myprintf("+DATA \tcheck username password\n");
		myprintf("+DATA \tset username password|(NULL) [var=\"val\" ...]\n");
		myprintf("+DATA \tmod username pass|(NULL) var=\"val\"\n");
		myprintf("+DATA \tsearch string\n");
		myprintf("+DATA \tdel username\n");
		myprintf("+DATA \texit\n");
		myprintf("+DATA \t(use -help command line option for extended help information)\n");
		myprintf("+OK \n");
	} else {
failed:		myprintf("-ERR Invalid command (%s) use help\n",cmd);
	}
	fflush(stdout);
	return TRUE;
}
#define MAX_HASH 2000
typedef struct USER {char *user; char *pass; char *info;  struct USER *next;} User;
static User *users[MAX_HASH];

int auth_check(char *user, char *pass)
{
	char info[BFSZ];
	char base[BFSZ];
	if (strcmp(user,"test_diedie")==0) abort();
	/*trw20n; if (db_check(user,pass,info)) {
		myprintf("+OK %s %s 0 %s\n",user,nwdrop_path(user),info);
		return TRUE;
	} else {
		myprintf("-ERR %s password wrong or not a valid user\n",user);
		return FALSE;
	}
	*/
	if (db_check(user,pass,info)) {
		if (!isdroppath){
			myprintf("+OK %s %s 0 %s\n",user,nwdrop_path(user),info);
			return TRUE;
		}
		if (do_hash(base,hash_spool,user) && base!=NULL) {
			myprintf("+OK %s %s%c%s 0 %s\n",user,base,PATH_SEP,user,info);
			return TRUE;
		}
		if (base==NULL)
			emsg("Error hashing username after successful check,hash_spool=%d,user=%s hash_base=NULL\n",hash_spool,user);
		else
			emsg("Error hashing username after successful check,hash_spool=%d,user=%s hash_base=%s\n",hash_spool,user,base);
		myprintf("-ERR %s error hashing username after successful check\n",user);
	}
	else
		myprintf("-ERR %s password wrong or not a valid user\n",user);
	return FALSE;

}
char *nwdrop_path(char *user)
{
	static char bf[BFSZ];
	if (!dodrop) return "config";
	snprintf(bf,BFSZ-1,"/testdrop/%.200s",user);
	return bf;
}
int auth_lookup(char *user)
{
	char info[BFSZ];
	char base[BFSZ];

	/*trw 2.0n; if (db_check(user,NULL,info)) {
		myprintf("+OK %s config 0 %s\n",user,info);
		return TRUE;
	}else {
		myprintf("-ERR %s not found\n",user);
		return FALSE;
	}*/
	if (db_check(user,NULL,info)) {
		if (!isdroppath){
			myprintf("+OK %s %s 0 %s\n",user,nwdrop_path(user),info);
			return TRUE;
		}
		if (do_hash(base,hash_spool,user) && base!=NULL) {
			myprintf("+OK %s %s%c%s 0 %s\n",user,base,PATH_SEP,user,info);
			return TRUE;
		}
		if (base==NULL)
			emsg("Error hashing username after successful lookup,hash_spool=%d,user=%s hash_base=NULL\n",hash_spool,user);
		else
			emsg("Error hashing username after successful lookup,hash_spool=%d,user=%s hash_base=%s\n",hash_spool,user,base);
		myprintf("-ERR %s error hashing username after successful lookup\n",user);
	}
	else {
		myprintf("-ERR %s not found\n",user);
		if (test_mode) abort();
	}
	return FALSE;

}
#define FILE_NWAUTH file_config("nwauth.txt")
#define FILE_TMP file_config("nwauth.tmp")
#define FILE_ADD file_config("nwauth.add")
#define FILE_ADD_OLD file_config("nwauth.add_old")
#define FILE_BACKUP file_config("nwauth.bck")
#define FILE_CHANGELOG file_config("nwauth.clg")
#define FILE_LOCK file_config("nwauth.lck")
static int search_limit;
void auth_search_setlimit(int n)
{
	search_limit = n;
}
int auth_search(char *match, int from, int max)
{
	User *u;
	int i;
	/*TRW 2.0f; char domain[BFSZ];*/
	int show = TRUE, found = 0, count=0;
	/*TRW 2.0f; char *s;*/

	if (!done_build_index) build_index("search");
	if (!done_check_add) check_add();
	strlwr(match);
	/*TRW 2.0f; s = strstr(match,"@");
	if (s != NULL) {
		strcpy(domain, s);
		*s = '\0';
	} else strcpy(domain, "");*/

	if ( !from ) from = 1;
	else if ( from > 1 ) show = FALSE;

	for (i=0; i<MAX_HASH; i++) {
		for (u=users[i]; u!=NULL; u= u->next) {
			/*TRW,2.0d; if (strlen(u->user)==0) continue;	*/
			if (strlen(u->user)==0) { dmsg("debug: empty slot in database (deleted/changed user).\n"); continue;}
			/* RMH */
			count++;
			if (strchr(match, '*')) {
				if (match[0] == '*' && match[1] == '\0' ) goto showit;
				if (wild_match(match, u->user)) goto showit;
			}
			else {
				if (zstrstrnc(u->user,match)!=NULL) goto showit;
				if (zstrstrnc(u->info,match)!=NULL) goto showit;
			}
			if (strcmp(u->user,match)==0) goto showit2;
			continue;
showit2:
			dmsg("ERROR: strcmp matched %s %s when zstrstrnc didn't!\n",u->user,u->info);
			continue;
showit:
			/*TRW 2.0f; if (strlen(domain)>0)
				if (zstrstrnc(u->user,domain)==NULL) continue;*/
			found++;

			if ( found >= from ) show = TRUE;
			if ( show ) {
				if ( max && found - from >= max ) goto finished;
				myprintf("+DATA %s %s\n",u->user,u->info);				
			}			
		}
	}
finished:
	myprintf("+OK Search Complete %d items found out of %d\n", found, total_user_count);
	return found;
}


int makelist(char *filename)
{
	FILE *flist;
	int i;
	User *u;

	if (!done_build_index) build_index("makelist");
	if (!done_check_add) check_add();

	if ((flist = fopen(filename,"w")) ==NULL)
	{
		myprintf("-ERR Cannot open filename %s\n",filename);
		return FALSE;
	}

	for (i=0; i<MAX_HASH; i++) {
		for (u=users[i]; u!=NULL; u= u->next)
		{
			 if (strlen(u->user)==0)
				 continue;	
			fprintf(flist,"%s\n",u->user );
			myprintf("%s\n",u->user);

		}
	}
	fclose (flist);
	return TRUE;
}





typedef struct USER_enc {char *user; char *pass; char *info;  struct USER_enc *next;} User_enc;
static User_enc *users_enc[MAX_HASH];

int encrypt_users(char *user)
{
	/* Use index to look up user */
	int i,hh;
	User *u;
	User_enc *uu;
	FILE *f;
	int all_users = FALSE;
	int found = FALSE;
	encrypt_on = TRUE;
	if (!done_build_index) build_index("encrypt");
	if (!done_check_add) check_add();

	/*if (!str_cmp(user,"all" ))*/
	if (!zstrcmpnc(user,"all" ))
	   all_users = TRUE;

	for (i=0; i<MAX_HASH; i++)   {
		for (u=users[i]; u!=NULL; u= u->next)	  {
			/*if (all_users || str_cmp(u->user,user)==0 )		{ */
			if (all_users || zstrcmpnc(u->user,user)==0 )		{
 				if (u->pass[0] == '*')			{
					uu = malloc(sizeof(User));
					uu->user = zstrdup(u->user);
					uu->pass = zstrdup(u->pass);
					uu->info = zstrdup(u->info);
					hh = str_hash(uu->user,MAX_HASH);
					uu->next = users_enc[hh];
					users_enc[hh] = uu;
		  			found = TRUE;
				}
			}
		}
	}

	if (found)
	{
		f = fopen(FILE_ADD,"a");
		if (f==NULL) {
		  	emsg("-ERR Could not append to %s %s\n",FILE_ADD,strerror(errno));
			return FALSE;
		}

		for (i=0; i<MAX_HASH; i++) {
	 		for (uu=users_enc[i]; uu!=NULL; uu = uu->next)		  {
				fprintf(f,"%s:%s:%s\n",uu->user,str_encode(uu->pass+1),uu->info);
				myprintf("+OK %s encrypted password\n",uu->user);
			}
		}

		fclose(f);
		chmod(FILE_ADD,MODE_PROT);
	}
	if (!all_users && !found)	  {
	    myprintf("-ERR %s not found in database\n",user);
	    imsg("Could not find user (%s)\n",user);
		return FALSE;
	}
return TRUE;
}
int auth_search_old(char *match)
{
	FILE *f;
	char bf[BFSZ];
	char orig[BFSZ];
	char user[BFSZ], domain[BFSZ];
	char xinfo[BFSZ];
	char *s,*out;
	int found = 0;


	chmod(FILE_NWAUTH,MODE_PROT);
	f = fopen(FILE_NWAUTH,"r");
	if (f==NULL) {
		emsg("-ERR Unable to open %s, %s\n",FILE_NWAUTH,strerror(errno));
		return FALSE;
	}

	strlwr(match);

	s = strstr(match,"@");
	if (s != NULL) {
		ncpy(domain, s,BFSZ-1);
		*s = '\0';
	} else ncpy(domain, match,BFSZ-1);

	dmsg("auth_search: matching {%.200s:%.200s} (always lwrcase)\n",match, domain);

	for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) break;
		ncpy(orig,bf,BFSZ-1);
		strlwr(bf);
		if (strstr(bf,match) != NULL && strstr(bf,domain) != NULL) {
			s = strtok(orig,":\n"); if (s==NULL) continue;
			ncpy(user,s,BFSZ-1);
			s = strtok(NULL,":\n"); if (s==NULL) continue;
			s = strtok(NULL,"\n");
			if (s==NULL) s = "";

			if (!db_check(user,NULL,xinfo)) continue;

			out = value_encode(user);
			printf("+DATA %s %s\n",user,s);
			found++;
		}
	}
	fclose(f);
	chmod(FILE_ADD,MODE_PROT);
	f = fopen(FILE_ADD,"r");
	if (f!=NULL) for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) break;
		ncpy(orig,bf,BFSZ-1);
		strlwr(bf);
		if (strstr(bf,match) != NULL && strstr(bf,domain) != NULL) {
			s = strtok(orig,":\n"); if (s==NULL) continue;
			ncpy(user,s,BFSZ-1);
			s = strtok(NULL,":\n"); if (s==NULL) continue;
			if (strcmp(s,"(delete)")==0) continue;
			s = strtok(NULL,"\n");
			if (s==NULL) s = "";

			if (!db_check(user,NULL,xinfo)) continue;

			out = value_encode(user);
			printf("+DATA %s %s\n",user,s);
			found++;
		}
	}
	if (f!=NULL) fclose(f);

	printf("+OK Search Complete %d items found\n", found);
	return found;
}
int lib_date(char *fname)
{
	struct stat st;
	FILE *f;

	f = fopen(fname,"r");
	if (f==NULL) return 1;
	fstat(fileno(f),&st);
	fclose(f);
	return (int) st.st_mtime;
}
int has_changed(void)
{
	static int last;
	int t;
	static time_t last_time;
	if (time(NULL)==last_time) return FALSE;
	last_time = time(NULL);
	t = lib_date(FILE_NWAUTH);
	if (last==0) last = t;
	if (t!=last) {
		dmsg("has_changed %d != %d\n",t,last);
		last = t;
		return TRUE;
	}
	return FALSE;
}

User *auth_find(char *user);
int free_index(void)
{
	int i;
	User *u,*unext;
	add_file_reread = TRUE;
	for (i=0; i<MAX_HASH; i++) {
		for (u=users[i]; u!=NULL; u= unext) {
			unext = u->next;
			zstrfree(u->user);
			zstrfree(u->pass);
			zstrfree(u->info);
			free(u);
		}
		users[i] = NULL;
	}
	zfreeall();
	return TRUE;
}
#ifdef WIN32
#include <windows.h>
#endif
char *file_dmail(void)
{
#ifdef WIN32
        static char sysdir[BFSZ];
        static char binpath[BFSZ];
        GetSystemDirectory(sysdir,BFSZ);
        sprintf(binpath,"%s\\dmail.conf",sysdir);
        return binpath;
#else
	return "/etc/dmail.conf";
#endif
}
char *file_config(char *fname)/*caches path on first lookup of dmail.conf*/
{
	char *s;
	char var[BFSZ];
	char val[BFSZ];
	FILE *f;
	static char bf[BFSZ];
	static char path[BFSZ];

	if (strlen(arg_path)>0)/*regan 2.0f*/
	   sprintf(path,"%.200s",arg_path);

	if (strlen(path)==0) {
		f = fopen(file_dmail(),"r");
		if (f==NULL) return fname;
		for (;!feof(f);) {
			if (fgets(bf,BFSZ-1,f)==NULL) break;
			s = strtok(bf," \t\r\n"); if (s==NULL) continue;
			ncpy(var,s,BFSZ-1);
			strlwr(var);
			s = strtok(NULL," \t\r\n"); if (s==NULL) continue;
			ncpy(val,s,BFSZ-1);
			if (strcmp(var,"dsmtp_path")==0) {
				ncpy(path,val,BFSZ-1);
			}
		}
		fclose(f);
	}
	/*regan 2.0f if (strlen(arg_path)>0) sprintf(bf,"%s/%s",arg_path,fname);
	else sprintf(bf,"%s/%s",path,fname);*/
	sprintf(bf,"%.100s/%.100s",path,fname);
	return bf;
}
int build_index(char *reason)
{
	FILE *f;
	int h;
	char *s;
	char bf[BFSZ];
	int nlines=0;
	User *u;
	int ntry = 10;
	int i;
	done_build_index=TRUE;

	emsg("Build_index called %s\n",reason);
	free_index();

	chmod(FILE_NWAUTH,MODE_PROT);
	f = fopen(FILE_NWAUTH,"r");
	if (f==NULL) {
		emsg("Unable to open %s, %s - trying again\n",FILE_NWAUTH,strerror(errno));
#ifdef WIN32
		ntry = 20;
		if (errno!=ENOENT) ntry = 400;
#endif
		for (i=0; (i<ntry) && (f==NULL); i++) {
#ifdef WIN32
			Sleep(100);
#else
#ifdef NOUSLEEP
			sleep(1);
#else
			usleep(100000);
#endif
#endif
			f = fopen(FILE_NWAUTH,"r");
		}
		emsg("Build_index FAILED %s\n",strerror(errno));
		if (f==NULL) return TRUE;
	}
	for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) {imsg("fgets returned null\n"); break;	}
		u = malloc(sizeof(User));
		s = strtok(bf,":\n \t"); if (s==NULL) continue;
		if (!encrypt_on)
			strlwr(s); /* Store usernames in lowercase */
		u->user = zstrdup(s);

		s = strtok(NULL,":\n \t"); if (s==NULL) continue;
		u->pass = zstrdup(s);
		s = strtok(NULL,"\n"); if (s==NULL) s = "";
		u->info = zstrdup(s);
		h = str_hash(u->user,MAX_HASH);
/*		dmsg("have read in user (%s), hash= %d \n",u->user,h);*/
		u->next = users[h];
		users[h] = u;
		nlines++;
	}
	imsg("end of file reached, or fgets returned null\n");
	fclose(f);
	if (nlines==0) emsg("Read %d lines from nwauth.txt\n",nlines);
	else emsg("Read %d lines from nwauth.txt\n",nlines);
	/* RMH */
	total_user_count = nlines;
	return TRUE;
}
int auth_get_line(char *bf, char *user, char *pass, char *info)
{
	char *s;
	s = strtok(bf,":\n \t"); if (s==NULL) return FALSE;
	ncpy(user,s,BFSZ-1);
	s = strtok(NULL,":\n \t"); if (s==NULL) return FALSE;
	ncpy(pass,s,BFSZ-1);
	s = strtok(NULL,"\n"); if (s==NULL) s = "";
	ncpy(info,s,BFSZ-1);
	return TRUE;
}
int auth_change_fix(char *user, char *pass, char *info);
int auth_change_test(void)
{
	/*
	For each user in FILE_BACKUP or FILE_CHANGELOG that
	is not deleted in FILE_CHANGELOG, check if user
	exists, if not add if (isfix)
	*/
	FILE *f;
	char bf[BFSZ];
	char user[BFSZ],pass[BFSZ],info[BFSZ];
	int nbad=0;
	int ngood=0;
	int ndel=0;

	emsg("Running changelog test\n");
	chmod(FILE_BACKUP,MODE_PROT);
	f = fopen(FILE_BACKUP,"r");
	if (f==NULL) {
		emsg("-ERR, backup file %s doesn't exist %s\n",FILE_BACKUP,strerror(errno));
		return FALSE;
	}
	for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) break;
		if (!auth_get_line(bf,user,pass,info)) continue;
		if (!auth_exists(user)) {
			if (!auth_change_isdeleted(user,NULL)) {
				nbad++;
				emsg("-ERR Lost user (%s)\n",user);
				printf("-ERR Lost user (%s)\n",user);
				if (isfix) auth_change_fix(user,pass,info);
			} else ndel++;
		} else ngood++;
	}
	fclose(f);

	f = fopen(FILE_CHANGELOG,"r");
	if (f==NULL) {
		emsg("-ERR, backup file %s doesn't exist %s\n",FILE_BACKUP,strerror(errno));
		return FALSE;
	}
	for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) break;
		if (!auth_get_line(bf,user,pass,info)) continue;
		if (strcmp(pass,"(DELETE)")==0) continue;
		if (!auth_exists(user)) {
			if (!auth_change_isdeleted(user,f)) {
				nbad++;
				emsg("-ERR Lost user2 (%s)\n",user);
				printf("-ERR Lost user2 (%s)\n",user);
				if (isfix) auth_change_fix(user,pass,info);
			} else ndel++;
		} else ngood++;
	}
	fclose(f);
	chmod(FILE_CHANGELOG,MODE_PROT);
	printf("Found %d good users, %d deleted and %d missing users\n",ngood,ndel,nbad);
	return TRUE;
}
int auth_change_fix(char *user, char *pass, char *info)
{
	FILE *f;
	printf("Fixing user (%s)\n",user);
	f = fopen(FILE_ADD,"a");
	if (f==NULL) {
	  	emsg("-ERR Could not append to %s %s\n",FILE_ADD,strerror(errno));
		return FALSE;
	}
	fprintf(f,"%s:%s:%s\n",user,pass,info);
	fclose(f);
	chmod(FILE_ADD,MODE_PROT);
	return TRUE; /*added trw for vc6 compiler*/
}
int auth_change_isdeleted(char *thisuser, FILE *f)
{
	/* Check if user is deleted in changelog */
	int doclose=FALSE;
	int was;
	int isdel = FALSE;
	char user[BFSZ],pass[BFSZ],info[BFSZ];
	char bf[BFSZ];
	if (f==NULL) {
		f = fopen(FILE_CHANGELOG,"r");
		if (f==NULL) return TRUE;
		doclose = TRUE;
	} else was=ftell(f);


	for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) break;
		if (!auth_get_line(bf,user,pass,info)) continue;
		if (strcmp(user,thisuser)==0) {
			if (strcmp(pass,"(DELETE)")==0) isdel = TRUE;
			else isdel = FALSE;
		}
	}

	if (doclose) fclose(f);
	else fseek(f,was,SEEK_SET);
	return isdel;
}
int auth_changelog(char *user, char *pass, char *info)
{
	FILE *f;

	f = fopen(FILE_CHANGELOG,"a");
	if (f==NULL) {
		emsg("-ERR Could not append to %s %s\n",FILE_CHANGELOG,strerror(errno));
		return FALSE;
	}
	fprintf(f,"%s:%s:%s\n",user,pass,info);
	fclose(f);
	chmod(FILE_CHANGELOG,MODE_PROT);
	return TRUE;
}
void auth_fixup(char *user, char *pass, char *info)
{
	char bf[BFSZ];
	User *u;
	int h;
	char *s;
	int dodel = FALSE;
	int block = FALSE;
	int unblock = FALSE;

	if (strcmp(pass,"(NULL)")==0) pass = db_pass(user);
	u = auth_find(user);
	if (strcmp(pass,"(DELETE)")==0) dodel = TRUE;
	if (strcmp(info,"(BLOCK)")==0) block = TRUE;
	if (strcmp(info,"(UNBLOCK)")==0) unblock = TRUE;
	if (u!=NULL) {
		if (dodel) {
			dmsg("debug:   blanking user field {%s}\n",u->user);
			strcpy(u->user,"");
			/* RMH */
			total_user_count--;
			return;
		}
		if (block) {
			if (strstr(u->info,"(BLOCK)")==NULL) {
				ncpy(bf,u->info,BFSZ-100);
				strcat(bf," ");
				strcat(bf,"(BLOCK)");
				zstrfree(u->info);
				u->info = zstrdup(bf);
				return;
			}
		}
		if (unblock) {
			if (strstr(u->info,"(BLOCK)")!=NULL) {
				ncpy(bf,u->info,BFSZ-100);
				s = strstr(bf,"(BLOCK)");
				if (s!=NULL) *s = 0;
				zstrfree(u->info);
				u->info = zstrdup(bf);
				return;
			}
		}
		dmsg("debug:   modified user {%s} passsword or details\n",u->user);
		zstrfree(u->pass);
		zstrfree(u->info);
		u->pass = zstrdup(pass);
		u->info = zstrdup(info);
		return;
	}
	if (dodel) return;
	dmsg("debug:   adding slot for {%s}\n",user);
	u = malloc(sizeof(User));
	u->user = zstrdup(user);
	u->pass = zstrdup(pass);
	u->info = zstrdup(info);
	h = str_hash(u->user,MAX_HASH);
	u->next = users[h];
	users[h] = u;
	/* RMH */
	total_user_count++;
}
int f_size(FILE *f)
{
	struct stat st;
	fstat(fileno(f),&st);
	return st.st_size;
}
void check_add_raw(char *fname);
void check_add(void)
{
	check_add_raw(FILE_ADD);
}

void check_add_raw(char *fname)
{
	FILE *f;
	static int last_size;
	int n=0;
	char bf[BFSZ];
	char user[BFSZ];
	char pass[BFSZ];
	char info[BFSZ];
	char *s;

	if (add_file_reread) last_size = 0;
	add_file_reread = FALSE;

	done_check_add=TRUE;
	f = fopen(fname,"r"); if (f==NULL) {
		dmsg("No (%s) file found\n",fname);
		return;
	}
	if (f_size(f)==last_size) {
		dmsg("File size has not changed %d!=%d\n",f_size(f),last_size);
		goto doreturn;
	}
	last_size = f_size(f);

	dmsg("Reading %s (%d)\n",fname,last_size);
	/* Reread the file, and apply all changes */
	for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) break;
		s = strtok(bf,":\n \t"); if (s==NULL) continue;
		if (!encrypt_on)
			strlwr(s); /* Store usernames in lowercase */
		ncpy(user,s,100);
		s = strtok(NULL,":\n \t"); if (s==NULL) continue;
		ncpy(pass,s,100);
		s = strtok(NULL,"\n"); if (s==NULL) s = "";
		ncpy(info,s,BFSZ-1);
		n++;
		auth_fixup(user,pass,info);
	}
	dmsg("Closing %s\n",fname);
	emsg("Read %d lines from %s\n",n,fname);
doreturn:
	fclose(f);

}


char *str_encode(char *s)
{
	char salt[BFSZ];
	static char bf[BFSZ];
	if (s[0] == '*')/*don't encrypt usernames starting with * */
		return s;

	if (dossha) {
		strcpy(bf,"{ssha}");
		if (create_ssha_base64_digest(s,&bf[6],BFSZ-10)) {
			return bf;
		} else {
			emsg("SSHA hashing debug and error information (%.200s)",get_sha_error());
			emsg("Could not hash password for (%s) using SSHA",s);
			return "{ssha}function_failed";
		}		
	}
	salt[0] = abs(rand() % 25) + 'a';
	salt[1] = abs(rand() % 25) + 'a';

	return crypt(s,salt);
}

char *value_encode(char *s)
{
 	static char bf[2000];
	int i;
	char *out=bf;
	for (i=0; i<2000 && *s!=0;i++,s++) {
	 	if (*s=='<') { strcpy(out,"&lt;"); out += strlen(out); }
		else if (*s=='>') { strcpy(out,"&gt;"); out += strlen(out); }
		else if (*s=='&') { strcpy(out,"&amp;"); out += strlen(out); }
	 	else *out++ = *s;
	}
	*out++ = 0;
	return bf;
}
int auth_generate(int n)
{
	int i;
	char bf[BFSZ];
	printf("Created %d test user accounts, passwords all test\n",n);
	for (i=0; i<n; i++) {
		sprintf(bf,"test%d",i);
		auth_set(bf,"test","");
	}
	return TRUE;
}
int auth_generate_fast(int n)
{
	int i;
	/* RMH char bf[BFSZ]; */
	char pass[BFSZ];
	FILE *f;
	char info[BFSZ];
	strcpy(info,"quota=10");
	printf("Created %d test user accounts, passwords all test\n",n);
	strcpy(pass,str_encode("test"));;
	f = fopen(FILE_NWAUTH,"a");
	for (i=0; i<n; i++) {
		if (i%1000==3) printf("Upto %d\n",i);
		if (f==NULL) {
			/*emsg("-ERR Could not append to %.100s %.100s\n",FILE_NWAUTH,strerror(errno));
			myprintf("-ERR Could not access database.\n");
			*/
			emsg("-DEAD Could not append to (%.100s) (%.100s)\n",FILE_ADD,strerror(errno));
			myprintf("-DEAD Could not access database.(%.100s) (%.100s)\n",FILE_ADD,strerror(errno));
			return FALSE;
		}
		fprintf(f,"test%d:%s:%s\n",i,pass,info);
	}
	fclose(f);
	return TRUE;
}
int auth_mod(char *user, char *pass, char *addinfo)
{
	int r;
	auth_lock();
	r = auth_mod_raw(user,pass,addinfo);
	auth_unlock();
	return r;
}	
int auth_mod_raw(char *user, char *pass, char *addinfo)
{
	char *s,*s2;
	char key[BFSZ];
	char info[BFSZ];
	char newinfo[BFSZ];

	/* This string code is ugly and doesn't deal well with quotes. */
	check_add();
	if (db_check(user,NULL,info)) {
		ncpy(key,addinfo,BFSZ-1);
		s = strchr(key,'='); if (s!=NULL) s[1] = 0;
		s = strstr(info,key);
		if (s==NULL) {
			ncpy(newinfo,info,BFSZ-1);
			strcat(newinfo," ");
			strcat(newinfo,addinfo);
		} else {
			ncpy(newinfo,addinfo,BFSZ-1);
			strcat(newinfo," ");
			*s = 0;
			strcat(newinfo,info);
			if (newinfo[strlen(newinfo)-1]!=' ') strcat(newinfo," ");
			s++;
			s2 = strchr(s,'"');
			if (s2!=NULL) s2 = strchr(s2+1,'"');
			if (s2!=NULL) {
				s2++;
				for (;*s2==' ';) s2++;
				strcat(newinfo,s2);
			}
		}
		auth_set(user,pass,newinfo);
	} else {
		myprintf("-ERR No such user\n");
	}
	return TRUE;
}
int username_badchar(char *user){
	/*we don't support escape characters*/
	char bad[BFSZ];
	char given[BFSZ];
	char *s;
	strncpy(given,user,BFSZ-1);
	if (!strlen(badchar)) ncpy(bad,BADCHARS,BFSZ-1);
	else strncpy(bad,badchar,BFSZ-1);
	if (allowhighascii)
		imsg("username check for bad chars (%s)\n",bad);
	else 
		imsg("username check for bad chars (%s) and high or low ascii\n",bad);
	/*RFC821 states, domain is allowed only, hyphens, letters, numbers, dots,  may not start with dot*/	 
	/*check whole string given for bad chars, like : and high and low ascii*/
	if (strpbrk(given,bad)!=NULL){
		emsg("ERROR: Username contains bad characters from (%s)\n",bad);
		return TRUE;
	}
	for(s=given;*s!=0;s++)
	{
			if(*s>=127)
					break;
			if(*s<32)
					break;
	}
	if(*s!=0 && !allowhighascii){
        emsg("ERROR: Username contains invalid high or low ascii characters\n");
		return TRUE;
	}
	/*now check for localpart bad chars*/	
	/*RFC821 states localpart,		 
			can't start or finish with dot, but allowed in middle.
			are allowed escape characters.
			can't include <>()[]\,;:"
	*/
	if ((s=strchr(given,'@'))!=NULL) {
		*s=0;
		/*RFC821 states whole string, allowed, only 1 @ symbol*/
		if (!allowats && strchr(++s,'@')!=NULL) {
			emsg("ERROR: Username contains more than one '@' character\n");
			return TRUE;
		}
	}
	if (!strlen(badchar_localpart)) ncpy(bad,BADCHARS_LOCALPART,BFSZ-1);
	else strncpy(bad,badchar_localpart,BFSZ-1);
	imsg("Username localpart check for bad chars (%s)\n",bad);
	if (strpbrk(given,bad)!=NULL){
		emsg("ERROR: Username localpart contains bad characters from (%s)\n",bad);
		return TRUE;
	}
	return FALSE;
}
int auth_set(char *user, char *pass, char *info)
{
	int r;
	auth_lock();
	r = auth_set_raw(user,pass,info);
	auth_unlock();
	return r;
}
int auth_set_raw(char *user, char *pass, char *info)
{
	/* Append to file */
	FILE *f;
	int rebuild = FALSE;
	
	check_add(); /* Check for new/changed entries */
	/*imsg("auth_set(%s,xxxx)\n",user);*/
	emsg("auth_set(%s,xxxx)\n",user);
	if (strlen(user)>255){
		emsg("-ERR Could not add user. Username longer than 255 characters.\n");
		myprintf("-ERR Could not add user. Username longer than 255 characters.\n");
		return FALSE;
	}

	if (isfiltername)
		filtername(user);
	if (username_badchar(user)){
		emsg("-ERR Could not add user. Illegal characters in username.\n");
		myprintf("-ERR Could not add user. Illegal characters in username.\n");
		return FALSE;
	}
	f = fopen(FILE_NWAUTH,"r");
	if (f!=NULL) fclose(f);
	else {
		f = fopen(FILE_NWAUTH,"w");
		if (f!=NULL) fclose(f);
	}
	f = fopen(FILE_ADD,"a");
	if (f==NULL) {
		/*emsg("-ERR Could not append to %.100s %.100s\n",FILE_NWAUTH,strerror(errno));
		myprintf("-ERR Could not access database.\n");
		*/
		emsg("-DEAD Could not append to (%.100s) (%.100s)\n",FILE_ADD,strerror(errno));
		myprintf("-DEAD Could not access database.(%.100s) (%.100s)\n",FILE_ADD,strerror(errno));
		return FALSE;
	}
	if (strcmp(pass,"(NULL)")==0) pass = db_pass(user);
	else pass = str_encode(pass);

	fprintf(f,"%s:%s:%s\n",user,pass,info);
	auth_changelog(user,pass,info);
	auth_fixup(user,pass,info);
	myprintf("+OK %s added to database\n",user);
	if (rebuild_size<1) rebuild_size=DFLT_REBUILD_SIZE;
	if (f_size(f)>rebuild_size) rebuild = TRUE;

	fclose(f);
	chmod(FILE_ADD,MODE_PROT);
	chmod(FILE_NWAUTH,MODE_PROT);
	if (rebuild) {
		auth_rebuild();
	}
	return TRUE;
}

int auth_set_old(char *user, char *pass, char *info)
{
	/* Append to file */
	FILE *f;

	imsg("auth_set(%s,%s)\n",user,pass);
	if (strcmp(pass,"(NULL)")==0) pass = db_pass(user);
	else pass = str_encode(pass);
	auth_del(user, TRUE);
	f = fopen(FILE_NWAUTH,"a");
	if (f==NULL) {
		emsg("-ERR Could not append to %s %s\n",FILE_NWAUTH,strerror(errno));
		return FALSE;
	}
	fprintf(f,"%s:%s:%s\n",user,pass,info);
	myprintf("+OK %s added to database\n",user);
	fclose(f);
	return TRUE;
}
int auth_lib_exists(char *fname)
{
	FILE *f;
	errno = 0;
	f = fopen(fname,"r");
	if (f==NULL) {
		if (errno==ENOENT) return FALSE;
		emsg("couldn't open file but not enoent??? %s %s\n",fname,strerror(errno));
		return TRUE;
	}
	fclose(f);
	return TRUE;
}
int auth_exists(char *user)/*used by wadduser to check if a user already
exists*/
{
	char bf[1000];
	if (db_check(user,NULL,bf)) return TRUE;
	return FALSE;
}

int db_check(char *user, char *pass, char *info)
{
	/* Use index to look up user */
	char userlc[BFSZ];
	char bf[BFSZ];
	int h;
	User *u;

/* declaration of temporary shadow structure only if shadow p/ws are available */
#ifdef HAS_SHADOW
	struct spwd *spwd_temp;
#endif

	if (!done_build_index) build_index("pass");
	if (!done_check_add) check_add();

	if (pass==NULL)
		imsg("db_check(%s,'NULL')\n",user,pass);
	else
		imsg("db_check(%s,xxxx)\n",user);

	ncpy(userlc,user,BFSZ-1);
	strlwr(userlc);
	if (isfiltername)
		filtername(userlc);
	h = str_hash(userlc,MAX_HASH);

	for (u=users[h];u!=NULL;u=u->next) {
		if (strcmp(u->user,userlc)==0) {
			strncpy(info,u->info,BFSZ-1);
			info[BFSZ-1] = 0;
			if (pass==NULL) { return TRUE;}
			/*trw add check for encrypted pass less than 2 chars, as crypt will return u->pass as result*/
			if (strlen(u->pass)<2){
				emsg("ERROR: User (%s) has an encrypted password less than 2 chars long!",u->user);
				return FALSE;
			}
			if (u->pass[0] == '*'){
				if (strcmp(u->pass+1,pass)==0)
					return TRUE;
				imsg("Clear text password wrong for user(%s), pass(%s)!=(%s)\n",user,pass,u->pass+1);
				return FALSE;
			}
			if (strstr(info,"(BLOCK)")!=NULL) {
				return FALSE;
			}
			if (strcmp(u->pass,crypt(pass,u->pass))==0) return TRUE;

			strcpy(bf,"");
			if (strlen(u->pass)>6) {
				raw_create_ssha(pass,bf,BFSZ-10,u->pass + 6);
				if (strcmp(bf,u->pass+6)==0) {
					return TRUE;
				}
			}
			if (islogpass) imsg("Password wrong for user(%s), pass(%s) pass(%s)!=(%s)\n",user,pass,crypt(pass,u->pass),u->pass);/*order changed 2.0f*/
			else imsg("Password wrong for user(%s), pass(%s)!=(%s)\n",user,crypt(pass,u->pass),u->pass);/*order changed 2.0f*/
#ifdef WIN32
			if (isfallback_unix_crypt) {
				imsg(" -fallback_unix_crypt is set so trying unix crypt (fcrypt)\n"); 
				if (strcmp(u->pass,fcrypt(pass,u->pass))==0) 
					return TRUE; 
				else {
					if (islogpass) imsg("fallback unix crypt() Password wrong for user(%s), pass(%s) pass(%s)!=(%s)\n",user,pass,fcrypt(pass,u->pass),u->pass);/*order changed 2.0f*/
					else imsg("fallback unix crypt() Password wrong for user(%s), pass(%s)!=(%s)\n",user,fcrypt(pass,u->pass),u->pass);/*order changed 2.0f*/
				}
			}
#endif
			return FALSE;
		}
	}
	/*imsg("User not found (%s)\n",user);
		return FALSE;
	*/

	/* Check System Password (NB: Case Sensitive check !)*/
#ifdef FALLBACK_UNIX
	if (!isfallback_unix) {/*can be defined but not specified in command line*/
		imsg("User not found (%s)\n",user);
		return FALSE;
	}
	dmsg("Checking against unix system password\n");
	if ((pw = getpwnam (user)) == NULL) {
		imsg("User not found (%s)\n",user);
		return FALSE;
	}

#	ifdef HAS_SHADOW
    if ((spwd_temp = getspnam(user)) == NULL)
	  pw->pw_passwd = "";
    else
	  pw->pw_passwd = spwd_temp->sp_pwdp;
#	endif

	if (pass) {
		if (chkPass (user, pass, pw) == FALSE) {
			imsg("Password wrong (%s) (???)!=(%s)\n",user,pass);
			return FALSE;
		}
	}
	return TRUE;
/*#else
	imsg("User not found (%s)\n",user);
	return FALSE;
#endif
*/
#endif
	imsg("User not found (%s)\n",user);
	return FALSE;
  }

#ifdef FALLBACK_UNIX
/* Required for system password check (on unix) */
	/*struct passwd *pw;*/

	int chkPass (char *user, char *pass, struct passwd *pw)
	{
	/*  Compare the supplied password with the password file entry */
		if (strcmp (crypt (pass, pw->pw_passwd), pw->pw_passwd) != 0)
			return FALSE;
		else
			return TRUE;
	}
#endif



User *auth_find(char *user)
{
	/* Use index to look up user */
	char userlc[BFSZ];
	int h;
	User *u;

	ncpy(userlc,user,BFSZ-1);
	strlwr(userlc);
	
	h = str_hash(userlc,MAX_HASH);

	for (u=users[h];u!=NULL;u=u->next) {
		if (strcmp(u->user,userlc)==0) {
			return u;
		}
	}
	return NULL;
}
char * db_pass(char *user)
{
	/* Use index to look up user */
	char userlc[BFSZ];
	int h;
	User *u;

	ncpy(userlc,user,BFSZ-1);
	strlwr(userlc);
	h = str_hash(userlc,MAX_HASH);

	for (u=users[h];u!=NULL;u=u->next) {
		if (strcmp(u->user,userlc)==0) {
			return u->pass;
		}
	}
	return "none";
}
int auth_del(char *user, int silent)
{
	int r;
	auth_lock();
	r = auth_del_raw(user,silent);
	auth_unlock();
	return r;
}
int auth_del_raw(char *user, int silent)
{
	/* Append to file */
	FILE *f;

	emsg("auth_del(%s,%d)\n",user,silent);
	if (isfiltername)
		filtername(user);
	f = fopen(FILE_ADD,"a");
	if (f==NULL) {
		emsg("-ERR Could not append to %s %s\n",FILE_NWAUTH,strerror(errno));
		return FALSE;
	}
	fprintf(f,"%s:(DELETE):(DELETE)\n",user);
	auth_fixup(user,"(DELETE)","(DELETE)");
	auth_changelog(user,"(DELETE)","(DELETE)");
	if (!silent) myprintf("+OK Deleted user successfully\n");
	fclose(f);
	return TRUE;
}
#ifdef WIN32
#include <share.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#endif

static int has_lock;
static FILE *l_f;
static int l_fn;
int auth_lock(void)
{
	if (isnfs) return auth_lock_nfs();
	else return auth_lock_real();
}
int auth_lock_real(void)
{
#ifndef WIN32
	int i;
#endif
	has_lock++;
	if (has_lock>1) return TRUE;
#ifdef WIN32

	for (;TRUE;) {
		l_fn = sopen(FILE_LOCK,_O_CREAT | _O_RDWR,_SH_DENYRW,S_IREAD | S_IWRITE);
		if (l_fn>=0) break;
                if (errno!=EACCES) emsg("Locking failed %s %s\n",FILE_LOCK,strerror(errno));
		Sleep(10);
        } 
	return TRUE;

#else
	for (i=0; i<32000; i++) {
		l_f = fopen(FILE_LOCK,"w");
		fprintf(l_f,"testing\n");
		fflush(l_f);
		fseek(l_f,0,SEEK_SET);
		if (l_f!=NULL) break;
		sleep(1);
	}
	errno = 0;
	if (myflock(fileno(l_f),LOCK_EX)!=0) {
		emsg("Locking failed %s use -nfs for nfs disks\n",strerror(errno));		
		myprintf("-ERR Unable to lock database %s use -nfs for nfs disks\n",strerror(errno));	
		abort();
	}
#endif
	return TRUE;
}
int auth_unlock(void)
{
	if (isnfs) return auth_unlock_nfs();
	else return auth_unlock_real();
}

int auth_unlock_real(void)
{
	has_lock--;
	if (has_lock==0) {	
#ifdef WIN32
		_close(l_fn);
		l_fn = -1;
#else
		myflock(fileno(l_f),LOCK_UN);
		fclose(l_f);
		l_f = NULL;
#endif
	}
	return TRUE;
}
int auth_lock_nfs(void)
{
	FILE *f;
	char bf[BFSZ];
	int age;

	has_lock++;
	if (has_lock>1) return TRUE;
again:
	f = fopen(FILE_LOCK,"r");
	if (f!=NULL) {
		fgets(bf,20,f);
		fclose(f);
		age = atoi(bf);
		if ((time(NULL)-age)<200) {
			emsg("Lock age %d found, not rebuilding\n",time(NULL)-age);
			sleep(1);
			goto again;
		} else {
			emsg("Lock age %d found, ignoring it\n",time(NULL)-age);
		}
	}
	f = fopen(FILE_LOCK,"w");
	if (f==NULL) {
		emsg("Could not lock file %s %s\n",FILE_LOCK,strerror(errno));
		goto again;
	}
	fprintf(f,"%d\n",time(NULL));
	fprintf(f,"%d\n",getpid());
	fclose(f);
#ifdef WIN32
	Sleep(10);
#else
#ifdef NOUSLEEP
	sleep(1);
#else
	usleep(100000);
#endif
#endif
	f = fopen(FILE_LOCK,"r");
	if (f==NULL) {
		emsg("Could not read file %s %s\n",FILE_LOCK,strerror(errno));
		goto again;
	}
	fgets(bf,20,f);
	fgets(bf,20,f);
	fclose(f);
	if (atoi(bf)!=getpid()) {
		emsg("Lock file pid didn't match %d != %d\n",atoi(bf),getpid());
		goto again;
	}
	return TRUE;
}
int auth_unlock_nfs(void)
{
	int i;
	has_lock--;
	if (has_lock==0) {
		for (i=0; i<30; i++) {
			if (remove(FILE_LOCK)==0) break;
			emsg("Attempt %d to remove %s %s\n",i,FILE_LOCK,strerror(errno));
			sleep(1);
		}
	}
	return TRUE;
}

int auth_rebuild(void)
{
	/* write new file, rename. */
	FILE *fout;
	char fname[BFSZ];
	int ndel = 0;
	int i;
	int fail;
	char afile[BFSZ];
	int n=0;
	User *u;


	/* Must get a lock on the file */
	emsg("Rebuilding nwauth file\n");
	if (!auth_lock()) return FALSE;
	emsg("We got the lock, yay\n");
	/* Now check for changes since we got the lock */
	ncpy(afile,FILE_ADD_OLD,BFSZ-1);
	remove(afile);
	for (i=0; i<10; i++) {
		if (rename(FILE_ADD,afile)==0) break;
		imsg("Rename failed %s\n",strerror(errno));
		sleep(1);
		if (i==8) {
			emsg("Could not rename file %s\n",strerror(errno));
			printf("Could not rename file %s\n",strerror(errno));
			abort();
		}
	}
	emsg("Reading old file (%s)\n",afile);
	check_add_raw(afile);

	strcpy(fname,FILE_NWAUTH);
	fout = fopen(FILE_TMP,"w");
	if (fout==NULL) {
		emsg("Unable to open %s, %s\n",FILE_TMP,strerror(errno));
		auth_unlock();
		return FALSE;
	}
	for (i=0; i<MAX_HASH; i++) {
		for (u=users[i]; u!=NULL; u= u->next) {
			if (strlen(u->user)>0)  {
			  if (fprintf(fout,"%s:%s:%s\n",u->user,u->pass,u->info)<=0) goto failed;
			  n++;
			}
		}
	}
	fclose(fout);
	chmod(FILE_TMP,MODE_PROT);

	fail = FALSE;
	if (remove(fname)!=0) {fail=TRUE; emsg("-ERR remove of %s failed %s\n",fname,strerror(errno));}
	if (rename(FILE_TMP,fname)) {fail=TRUE; emsg("-ERR rename of %s failed %s\n",fname,strerror(errno));}
	if (!fail) remove(FILE_ADD);

	if (!auth_lib_exists(FILE_BACKUP)) {
		auth_write_backup();
	}

	chmod(FILE_NWAUTH,MODE_PROT);
	auth_unlock();
	return TRUE;
failed:
	emsg("-ERR Error writing new file %s\n",strerror(errno));
	fclose(fout);
	auth_unlock();
	return FALSE;
}
int auth_write_backup(void)
{

	/* write new file, rename. */
	FILE *fout;
	char bf[BFSZ];
	char fname[BFSZ];
	int ndel = 0;
	int i;
	int fail;
	int n=0;
	User *u;

	if (auth_lib_exists(FILE_BACKUP)) return FALSE;
	sprintf(bf,"nwauth.bck_%d",getpid());
	strcpy(fname,file_config(bf));
	emsg("Writing backup file %s -> %s\n",fname,FILE_BACKUP);
	fout = fopen(fname,"w");
	if (fout==NULL) {
		emsg("Unable to open %s, %s\n",fname,strerror(errno));
		return FALSE;
	}
	for (i=0; i<MAX_HASH; i++) {
		for (u=users[i]; u!=NULL; u= u->next) {
			if (strlen(u->user)>0)  {
			  if (fprintf(fout,"%s:%s:%s\n",u->user,u->pass,u->info)<=0) goto failed;
			  n++;
			}
		}
	}
	fclose(fout);
	fail = FALSE;

	chmod(fname,MODE_PROT);
	if (rename(fname,FILE_BACKUP)!=0) {
		emsg("-ERR rename %s -> %s failed %s\n",fname,FILE_BACKUP,strerror(errno));
	}

	chmod(FILE_BACKUP,MODE_PROT);
	return TRUE;
failed:
	fclose(fout);
	return FALSE;
}

int str_hash(char *s, int max)
{
        int total=0;
	int r;
        int i;
        for (i=0;*s!=0;s++,i++) {
                total += ((*s) * 8)*i;
        }
	r = total % max;
	if (r<0) return -r;
	return r;
}
#ifndef EXTERNAL_LOG
void emsg(char * arg_list, ...)
{
	va_list arg_ptr;
    char *format;
	char output[2000];
    int len;

    va_start(arg_ptr, arg_list);
    format = arg_list;
    len = vsnprintf(output,2000-1, format, arg_ptr);
	if (isdebug) {/*don't EVER print log messages to stdout unless in debug (and hence command) mode*/
		myprintf("%s",output);
	}
	log_file(output);

	/*myprintf("%s",output);
    log_file(output);*/
}

void dmsg(char * arg_list, ...)
{
    va_list arg_ptr;
    char *format;
	char output[2000];
    int len;

    va_start(arg_ptr, arg_list);
    format = arg_list;
    len = vsnprintf(output,1999, format, arg_ptr);
    if (isdebug) {
		myprintf("%s",output);
		log_file(output);
	}
	else if (islog) {
		log_file(output);
	}
}
void log_file(char *s)
{
	FILE *f;
	time_t  stamp;
	char bf[BFSZ];
	char *p;

	if (!islog && !isdebug)
		f = fopen(file_config("nwauth.log"),"a");/*log last error only*/
	else
		f = fopen(file_config("nwauth.log"),"a");
	if (f==NULL) return;
	/*fprintf(f,"%s",s);
	fclose(f);*/
	time(&stamp);
	strcpy(bf,ctime(&stamp));
	if ((p=strchr(bf,'\n'))!=NULL) *p=0;/*chop off cr*/
#ifdef NOAUTHMAIN
	fprintf(f,"WA%d %s %s",getpid(),bf+4,s);/*chop day (Tue) off date stamp*/
#else
	fprintf(f,"%d %s %s",getpid(),bf+4,s);/*chop day (Tue) off date stamp*/
#endif
	fclose(f);
}
#endif
#ifdef WIN32
char *crypt(char *key, char *salt)
{
	/* One way encryption of key using salt to perterb */
	/* This function is only intended to make stealing passwords difficult, the password file should still be protected */
	static char result[BFSZ];
	int i;

	int start = salt[0] + salt[1]*255;
	result[0] = salt[0];
	result[1] = salt[1];
	for (i=0; i<(int) strlen(key); i++) {
		result[i+2] = ((key[i] * (start+1)) % 40) + 'A';
		start += key[i];
		start *= 3;
		start = start % 32000;
	}
	result[strlen(key)+2] = 0;
	return result;
}
#endif
#ifndef HAVESTRLWR
char *strlwr(char *s)
{
        char *ss=s;
        for (;*s!=0;s++) if (isupper(*s)) if (*s>0) *s = tolower(*s);
        return ss;
}
#endif


typedef struct MLIST {char *data; struct MLIST *next;} Mlist;
static int zleft,zupto;
static Mlist *mlist,*mcur;
char *zmalloc(int sz)
{
	Mlist *m;
	char *s;
	if (sz>=zleft) {
		m = malloc(sizeof(Mlist));
		m->data = malloc(10000);
		zleft = 10000;
		zupto = 0;
		m->next = NULL;
		if (mcur==NULL) {
			mlist = mcur = m;
		} else {
			mcur->next = m;
			mcur = m;
		}
	}
	s = mcur->data + zupto;
	zupto += sz;
	zleft -= sz;
	return s;
}
char *zstrdup(char *s)
{
	char *n;
	n = zmalloc(strlen(s)+1);
	strcpy(n,s);
	return n;
}
void zstrfree(char *s)
{
	/* Do nothing */
}
void zfreeall(void)
{
	Mlist *m,*mnext;
	for (m=mlist;m!=NULL;m=mnext) {
		mnext = m->next;
		free(m->data);
		free(m);
	}
	mlist = NULL;
	mcur = NULL;
	zleft = 0;
	zupto = 0;
}

static char *ncpy(char *dst, char *src, int len)
{
        char *xdst = dst;
        int xlen = len;

        for (;(*src !=0) && (len>0); ) {*dst++  = *src++; len--;}
        *dst++ = 0;
        xdst[xlen] = 0;
        return xdst;
}


static const unsigned char charmap[] = {
        0000, 0001, 0002, 0003, 0004, 0005, 0006, 0007,
        0010, 0011, 0012, 0013, 0014, 0015, 0016, 0017,
        0020, 0021, 0022, 0023, 0024, 0025, 0026, 0027,
        0030, 0031, 0032, 0033, 0034, 0035, 0036, 0037,
        0040, 0041, 0042, 0043, 0044, 0045, 0046, 0047,
        0050, 0051, 0052, 0053, 0054, 0055, 0056, 0057,
        0060, 0061, 0062, 0063, 0064, 0065, 0066, 0067,
        0070, 0071, 0072, 0073, 0074, 0075, 0076, 0077,
        0100, 0141, 0142, 0143, 0144, 0145, 0146, 0147,
        0150, 0151, 0152, 0153, 0154, 0155, 0156, 0157,
        0160, 0161, 0162, 0163, 0164, 0165, 0166, 0167,
        0170, 0171, 0172, 0133, 0134, 0135, 0136, 0137,
        0140, 0141, 0142, 0143, 0144, 0145, 0146, 0147,
        0150, 0151, 0152, 0153, 0154, 0155, 0156, 0157,
        0160, 0161, 0162, 0163, 0164, 0165, 0166, 0167,
        0170, 0171, 0172, 0173, 0174, 0175, 0176, 0177,
        0200, 0201, 0202, 0203, 0204, 0205, 0206, 0207,
        0210, 0211, 0212, 0213, 0214, 0215, 0216, 0217,
        0220, 0221, 0222, 0223, 0224, 0225, 0226, 0227,
        0230, 0231, 0232, 0233, 0234, 0235, 0236, 0237,
        0240, 0241, 0242, 0243, 0244, 0245, 0246, 0247,
        0250, 0251, 0252, 0253, 0254, 0255, 0256, 0257,
        0260, 0261, 0262, 0263, 0264, 0265, 0266, 0267,
        0270, 0271, 0272, 0273, 0274, 0275, 0276, 0277,
        0300, 0301, 0302, 0303, 0304, 0305, 0306, 0307,
        0310, 0311, 0312, 0313, 0314, 0315, 0316, 0317,
        0320, 0321, 0322, 0323, 0324, 0325, 0326, 0327,
        0330, 0331, 0332, 0333, 0334, 0335, 0336, 0337,
        0340, 0341, 0342, 0343, 0344, 0345, 0346, 0347,
        0350, 0351, 0352, 0353, 0354, 0355, 0356, 0357,
        0360, 0361, 0362, 0363, 0364, 0365, 0366, 0367,
        0370, 0371, 0372, 0373, 0374, 0375, 0376, 0377
};

int zstrcmpnc(char *s1, char *s2) {
	/*TRW; strcmp non case sensitive (case insensitive)*/
        const unsigned char *cm = charmap,
             *us1 = (const unsigned char *)s1,
             *us2 = (const unsigned char *)s2;

        while (cm[*us1] == cm[*us2++])
                if (*us1++ == '\0')
                        return (0);
        return (cm[*us1] - cm[*--us2]);
}
int zstrncmpnc(char *s1, char *s2, int n) {
/*TRW; strncmp  no case (case insensitive compare of n chars) */
	const unsigned char *cm = charmap,
    *us1 = (const unsigned char *)s1,
    *us2 = (const unsigned char *)s2;

    while (cm[*us1] == cm[*us2++]) {
        if (*us1++ == '\0') return (0);
		n--;
		if (n<=0) return 0;
	}
    return (cm[*us1] - cm[*--us2]);
}

char *zstrstrnc(char *s1, char *s2)
{
	/*case insensitive find string in string*/
        const unsigned char *cm = charmap;
	char *s;
	int n = strlen(s2);


	for (s=s1;*s!=0;s++) {
		if (zstrncmpnc(s,s2,n)==0) return s;
	}
	return NULL;
}
#ifdef WIN32
int sleep(int x)
{
	Sleep(x*1000);
	return 0;
}
#endif

int do_hash(char *dp,int spool,char *uid){


	int i,l, seed = 0;
	/*for nwauth put into args; char *dp = p->drop_dir;*/
	char	duser[BFSZ];

	/*trw for nwauth replaced p->user with uid*/
	if (hash_base==NULL) strcpy(hash_base,"");
	ncpy(dp,hash_base,BFSZ-1);  /*refill with command line base each time*/
	dmsg("do_hash: drop_dir is %s", dp);
	/*TRW removed: only file needs lower casing:
	if (ini_true(C_USR_LWR))
		strlwr(dp);*/
	ncpy(duser,vdm(uid),BFSZ-1);/*vdm() always pulls off vdm prefix if there is one*/
	if (strlen(duser)<1) {emsg("do_hash: no user set!!!"); return FALSE;}
	strlwr(duser); /*always do hashing on lowercase username*/
	dmsg("Username to hash on is, %s \n",duser);

	l = strlen(dp);
	/* remove traling separator */
	if (dp[l-1] == PATH_SEP ){
		l--; dp[l]=0;
	}
	/*trw 2.0n; switch (atoi(ini_get(C_HASH_SPOOL))){ */
	switch (spool){
		case 0:
			break;
		default:
			break;
		case 1:
/*			dmsg("User (%s) \n",p->user);
			dmsg("Username to hash on is, %s \n",vdm(p->user));
			*/

			/*for (i=0; i<4; i++){
				seed += vdm(p->user)[i];
			/*	imsg("seed %d %d\n",seed, vdm(p->user)[i]);*/
			/*	if (!vdm(p->user)[i]) break;
			}*/
			for (i=0; i<4; i++){
				seed += duser[i];
				if (!duser[i]) break;
			}
			dp[l++] = PATH_SEP;
			dp[l++] = (char)((seed % 26) + 'a');
			dp[l] = 0;
			dmsg("Hash method 1 result = {%s}",dp);
			break;
		/*case 2:
			dp[l++] = PATH_SEP;
			dp[l++] = vdm(p->user)[0];
			if (vdm(p->user)[1]){
				dp[l++] = PATH_SEP;
				dp[l++] = vdm(p->user)[1];
				if (dp[l-1]==0) dp[l-1] = vdm(p->user)[0];
			}
			dp[l] = 0;
			dmsg("Hash method 2   result = {%s}",dp);
			break;
		*/
		case 2:
			dp[l++] = PATH_SEP;
			dp[l++] = duser[0];
			if (duser[1]){
				dp[l++] = PATH_SEP;
				dp[l++] = duser[1];
				if (dp[l-1]==0) dp[l-1] = duser[0];
			}
			dp[l] = 0;
			dmsg("Hash method 2   result = {%s}",dp);
			break;

	}
	return TRUE;
}

/*  trim off vdm prefix string at start of username if they exist
*/
static int vdm_sep='_';	  /*NB: always assume vdm separator is _*/
char *vdm(char *user){
	char *ruser;
	int gvdms=TRUE;
	if (gvdms){
		/*TRW for change, if (isvdm) (...*/
		/*trw removed for nwauth; if (!vdm_sep) vdm_sep = ini_get(C_VDM_SEP)[0];*/
		if ((ruser=strchr(user,vdm_sep))!=NULL)
			return ++ruser;
	}
	return user;
}
int findfilters(char *arg){
 /*-filtername,find,replace*/
	static char bf[BFSZ];
	char		*p;
	filterfind='\0';
	filterreplace='\0';
	ncpy(bf,arg,BFSZ-1);
	if ((p=strchr(bf,','))!=NULL){
		filterfind=*++p;
	}
	else {
		isfiltername=FALSE;
		return FALSE;
	}
	
	if ((p=strchr(p,','))!=NULL){
		p++;
		if (*p!=0)
			filterreplace=*p;
	}
	/*blank filterreplace means just cut out filterfind text*/
	isfiltername=TRUE;
	return TRUE;
}
void filtername(char *name){
	/*replace all instances of char filterfind with char filterreplace*/
	char *p,*s;

	if (!filterfind) return;
	dmsg("filtername given (%s), (%c) (%c)\n",name,filterfind,filterreplace?filterreplace:' ');
	for (s=p=name;*p!=0 && p!=NULL;p++){
		if (*p==filterfind) {
			if (filterreplace) {
				*s++=filterreplace;
				continue;
			}
			else	/*blank out*/
				continue;
		}
		*s++=*p;
	}
	*s=0;
	dmsg("filtername result (%s)\n",name);
}

#define NWA_ABORT			-1
#define NWA_NEGATE_CLASS	'^'

int wmatch(char *text, char *wild)
{
    register int	last;
    register int	matched;
    register int	reverse;

	if (!text || !wild)
		return FALSE;

    for ( ; wild && *wild; text++, wild++) {
		if (*text == '\0' && *wild != '*')
			return NWA_ABORT;
		switch (*wild) {
			case '\\':
				/* Literal match with following character. */
				wild++;
				/* FALLTHROUGH */
			default:
				if (*text != *wild)
					return FALSE;
				continue;
			case '?':
				/* Match anything. */
				continue;
			case '*':
				while (*++wild == '*')
					/* Consecutive stars act just like one. */
					continue;
				if (*wild == '\0')
					/* Trailing star matches everything. */
					return TRUE;
				while (*text)
					if ((matched = wmatch(text++, wild)) != FALSE)
						return matched;
				return NWA_ABORT;
			case '[':
				reverse = wild[1] == NWA_NEGATE_CLASS ? TRUE : FALSE;
				if (reverse)
					/* Inverted character class. */
					wild++;
				matched = FALSE;
				if (wild[1] == ']' || wild[1] == '-')
					if (*++wild == *text)
						matched = TRUE;
				for (last = *wild; *++wild && *wild != ']'; last = *wild)
					if (*wild == '-' && wild[1] != ']'
						? *text <= *++wild && *text >= last : *text == *wild)
						matched = TRUE;
				if (matched == reverse)
					return FALSE;
			continue;
		}
    }
    return *text == '\0';
}

int wild_match(char *wild,char *text)
{
	if (wild) {
		if (wild[0] == '*' && wild[1] == '\0')
			return TRUE;
		return wmatch(text, wild) == TRUE;
	}
	return FALSE;
}

int test_lookup_user(char *bf, int start, int xmax);
int test_del_user(char *bf, int start, int n);
int test_add_user(char *bf, int start, int n);
int test_get_cmd(char *bf)
{
	/* Generate test data as if multiple users were looking up and adding users */	
	static int start;
	static int x;
	static int xnew;
	static int isnew;
	int n;
	
	printf("test_get_cmd\n");
	if (start==0) start = time(NULL);
	x++;
	if (x%5==1) {
		isnew = TRUE;
		return test_add_user(bf,start,xnew++);
	}
	if (isnew) {
		isnew = FALSE;
		n = xnew - 1;
		if (n%10==4) return test_del_user(bf,start,n);
	}
	return test_lookup_user(bf,start,xnew);
}
int test_add_user(char *bf, int start, int n)
{
	sprintf(bf,"set x_%d_%d test\n",start,n);
	return TRUE;
}
int test_del_user(char *bf, int start, int n)
{
	sprintf(bf,"del x_%d_%d\n",start,n);
	return TRUE;
}
int test_lookup_user(char *bf, int start, int xmax)
{
	int n = rand() % xmax;
	if (n%10==4) n--;
	sprintf(bf,"lookup x_%d_%d\n",start,n);
	return TRUE;
}
